@@ -10,13 +10,13 @@
  * ETHTOOL_PHYS_ID support by Chris Leech <christopher.leech@intel.com>
  * e1000 support by Scott Feldman <scott.feldman@intel.com>
  * e100 support by Wen Tao <wen-hwa.tao@intel.com>
+ * amd8111e support by Reeja John <reeja.john@amd.com>
  *
  * TODO:
  *   * no-args => summary of each device (mii-tool style)
  *   * better man page (steal from mii-tool?)
  *   * fall back on SIOCMII* ioctl()s and possibly SIOCDEVPRIVATE*
  *   * abstract ioctls to allow for fallback modes of data gathering
- *   * interface for EEPROM writing
  *   * symbolic names for msglvl bitmask
  */
 
@@ -52,6 +52,7 @@ static int do_sset(int fd, struct ifreq *ifr);
 static int do_gregs(int fd, struct ifreq *ifr);
 static int do_nway_rst(int fd, struct ifreq *ifr);
 static int do_geeprom(int fd, struct ifreq *ifr);
+static int do_seeprom(int fd, struct ifreq *ifr);
 static int do_test(int fd, struct ifreq *ifr);
 static int do_phys_id(int fd, struct ifreq *ifr);
 static int do_gpause(int fd, struct ifreq *ifr);
@@ -60,8 +61,8 @@ static int do_gring(int fd, struct ifreq *ifr);
 static int do_sring(int fd, struct ifreq *ifr);
 static int do_gcoalesce(int fd, struct ifreq *ifr);
 static int do_scoalesce(int fd, struct ifreq *ifr);
-static int do_gsum(int fd, struct ifreq *ifr);
-static int do_ssum(int fd, struct ifreq *ifr);
+static int do_goffload(int fd, struct ifreq *ifr);
+static int do_soffload(int fd, struct ifreq *ifr);
 static int do_gstats(int fd, struct ifreq *ifr);
 
 /* Syntax:
@@ -108,11 +109,16 @@ static int do_gstats(int fd, struct ifreq *ifr);
  *		[ raw on|off ] \
  *		[ offset N ] \
  *		[ len N ]
+ *	ethtool -E DEVNAME \
+ *		[ magic N ] \
+ *		[ offset N ] \
+ *		[ value N ] \
  *	ethtool -k DEVNAME
  *	ethtool -K DEVNAME \
  *		[ rx on|off ] \
  *		[ tx on|off ] \
- *		[ sg on|off ]
+ *		[ sg on|off ] \
+ *		[ tso on|off ]
  *	ethtool -r DEVNAME
  *	ethtool -p DEVNAME [ %d ]
  *	ethtool -t DEVNAME [ online|offline ]
@@ -125,6 +131,7 @@ static int do_gstats(int fd, struct ifreq *ifr);
  *		[ wol p|u|m|b|a|g|s|d... ] \
  *		[ sopass %x:%x:%x:%x:%x:%x ] \
  *		[ msglvl %d ]
+ *	ethtool -S DEVNAME
  */
 
 static void show_usage(int badarg)
@@ -174,11 +181,16 @@ static void show_usage(int badarg)
 		"		[ raw on|off ] \\\n"
 		"		[ offset N ] \\\n"
 		"		[ length N ]\n"
+ 		"	ethtool -E DEVNAME \\\n"
+		"		[ magic N ] \\\n"
+		"		[ offset N ] \\\n"
+		"		[ value N ]\n"
 		"	ethtool -k DEVNAME\n"
 		"	ethtool -K DEVNAME \\\n"
 		"		[ rx on|off ] \\\n"
 		"		[ tx on|off ] \\\n"
-		"		[ sg on|off ]\n"
+		"		[ sg on|off ] \\\n"
+		"		[ tso on|off ]\n"
 		"	ethtool -r DEVNAME\n"
 		"	ethtool -p DEVNAME [ %%d ]\n"
 		"	ethtool -t DEVNAME [online|(offline)]\n"
@@ -192,6 +204,7 @@ static void show_usage(int badarg)
 		"		[ wol p|u|m|b|a|g|s|d... ] \\\n"
 		"		[ sopass %%x:%%x:%%x:%%x:%%x:%%x ] \\\n"
 		"		[ msglvl %%d ] \n"
+		"	ethtool -S DEVNAME\n"
 	);
 	exit(badarg);
 }
@@ -204,6 +217,7 @@ static enum {
 	MODE_GREGS,
 	MODE_NWAY_RST,
 	MODE_GEEPROM,
+	MODE_SEEPROM,
 	MODE_TEST,
 	MODE_PHYS_ID,
 	MODE_GPAUSE,
@@ -212,15 +226,16 @@ static enum {
 	MODE_SCOALESCE,
 	MODE_GRING,
 	MODE_SRING,
-	MODE_GSUM,
-	MODE_SSUM,
+	MODE_GOFFLOAD,
+	MODE_SOFFLOAD,
 	MODE_GSTATS,
 } mode = MODE_GSET;
 
-static int gchecksum_changed = 0;
-static int csum_rx_wanted = -1;
-static int csum_tx_wanted = -1;
-static int csum_sg_wanted = -1;
+static int goffload_changed = 0;
+static int off_csum_rx_wanted = -1;
+static int off_csum_tx_wanted = -1;
+static int off_sg_wanted = -1;
+static int off_tso_wanted = -1;
 
 static struct ethtool_pauseparam epause;
 static int gpause_changed = 0;
@@ -266,6 +281,7 @@ static int port_wanted = -1;
 static int autoneg_wanted = -1;
 static int phyad_wanted = -1;
 static int xcvr_wanted = -1;
+static int advertising_wanted = -1;
 static int gset_changed = 0; /* did anything in GSET change? */
 static u32  wol_wanted = 0;
 static int wol_change = 0;
@@ -278,6 +294,10 @@ static int geeprom_changed = 0;
 static int geeprom_dump_raw = 0;
 static int geeprom_offset = 0;
 static int geeprom_length = -1;
+static int seeprom_changed = 0;
+static int seeprom_magic = 0;
+static int seeprom_offset = -1;
+static int seeprom_value = 0;
 static enum {
 	ONLINE=0,
 	OFFLINE,
@@ -302,10 +322,17 @@ static struct cmdline_info cmdline_geeprom[] = {
 	{ "raw", CMDL_BOOL, &geeprom_dump_raw, NULL },
 };
 
-static struct cmdline_info cmdline_checksum[] = {
-	{ "rx", CMDL_BOOL, &csum_rx_wanted, NULL },
-	{ "tx", CMDL_BOOL, &csum_tx_wanted, NULL },
-	{ "sg", CMDL_BOOL, &csum_sg_wanted, NULL },
+static struct cmdline_info cmdline_seeprom[] = {
+	{ "magic", CMDL_INT, &seeprom_magic, NULL },
+	{ "offset", CMDL_INT, &seeprom_offset, NULL },
+	{ "value", CMDL_INT, &seeprom_value, NULL },
+};
+
+static struct cmdline_info cmdline_offload[] = {
+	{ "rx", CMDL_BOOL, &off_csum_rx_wanted, NULL },
+	{ "tx", CMDL_BOOL, &off_csum_tx_wanted, NULL },
+	{ "sg", CMDL_BOOL, &off_sg_wanted, NULL },
+	{ "tso", CMDL_BOOL, &off_tso_wanted, NULL },
 };
 
 static struct cmdline_info cmdline_pause[] = {
@@ -404,15 +431,17 @@ static void parse_cmdline(int argc, char **argp)
 			else if (!strcmp(argp[i], "-G"))
 				mode = MODE_SRING;
 			else if (!strcmp(argp[i], "-k"))
-				mode = MODE_GSUM;
+				mode = MODE_GOFFLOAD;
 			else if (!strcmp(argp[i], "-K"))
-				mode = MODE_SSUM;
+				mode = MODE_SOFFLOAD;
 			else if (!strcmp(argp[i], "-i"))
 				mode = MODE_GDRV;
 			else if (!strcmp(argp[i], "-d"))
 				mode = MODE_GREGS;
 			else if (!strcmp(argp[i], "-e"))
 				mode = MODE_GEEPROM;
+			else if (!strcmp(argp[i], "-E"))
+				mode = MODE_SEEPROM;
 			else if (!strcmp(argp[i], "-r"))
 				mode = MODE_NWAY_RST;
 			else if (!strcmp(argp[i], "-p"))
@@ -433,14 +462,15 @@ static void parse_cmdline(int argc, char **argp)
 			    (mode == MODE_NWAY_RST) ||
 			    (mode == MODE_TEST) ||
 			    (mode == MODE_GEEPROM) ||
+			    (mode == MODE_SEEPROM) ||
 			    (mode == MODE_GPAUSE) ||
 			    (mode == MODE_SPAUSE) ||
 			    (mode == MODE_GCOALESCE) ||
 			    (mode == MODE_SCOALESCE) ||
 			    (mode == MODE_GRING) ||
 			    (mode == MODE_SRING) ||
-			    (mode == MODE_GSUM) ||
-			    (mode == MODE_SSUM) ||
+			    (mode == MODE_GOFFLOAD) ||
+			    (mode == MODE_SOFFLOAD) ||
 			    (mode == MODE_GSTATS) ||
 			    (mode == MODE_PHYS_ID)) {
 				devname = argp[i];
@@ -473,6 +503,14 @@ static void parse_cmdline(int argc, char **argp)
 				i = argc;
 				break;
 			}
+			if (mode == MODE_SEEPROM) {
+				parse_generic_cmdline(argc, argp, i,
+					&seeprom_changed,
+					cmdline_seeprom,
+					ARRAY_SIZE(cmdline_seeprom));
+				i = argc;
+				break;
+			}
 			if (mode == MODE_SPAUSE) {
 				parse_generic_cmdline(argc, argp, i,
 					&gpause_changed,
@@ -497,11 +535,11 @@ static void parse_cmdline(int argc, char **argp)
 				i = argc;
 				break;
 			}
-			if (mode == MODE_SSUM) {
+			if (mode == MODE_SOFFLOAD) {
 				parse_generic_cmdline(argc, argp, i,
-					&gchecksum_changed,
-			      		cmdline_checksum,
-			      		ARRAY_SIZE(cmdline_checksum));
+					&goffload_changed,
+			      		cmdline_offload,
+			      		ARRAY_SIZE(cmdline_offload));
 				i = argc;
 				break;
 			}
@@ -616,6 +654,28 @@ static void parse_cmdline(int argc, char **argp)
 			show_usage(1);
 		}
 	}
+
+	if (autoneg_wanted == AUTONEG_ENABLE){
+		if (speed_wanted == SPEED_10 && duplex_wanted == DUPLEX_HALF)
+			advertising_wanted = ADVERTISED_10baseT_Half;
+		else if (speed_wanted == SPEED_10 &&
+			 duplex_wanted == DUPLEX_FULL)
+			advertising_wanted = ADVERTISED_10baseT_Full;
+		else if (speed_wanted == SPEED_100 &&
+			 duplex_wanted == DUPLEX_HALF)
+			advertising_wanted = ADVERTISED_100baseT_Half;
+		else if (speed_wanted == SPEED_100 &&
+			 duplex_wanted == DUPLEX_FULL)
+			advertising_wanted = ADVERTISED_100baseT_Full;
+		else
+			/* auto negotiate without forcing */
+			advertising_wanted = ADVERTISED_100baseT_Full | 
+				ADVERTISED_100baseT_Half |
+				ADVERTISED_10baseT_Full | 
+				ADVERTISED_100baseT_Half;
+
+	}
+
 	if (devname == NULL) {
 		show_usage(1);
 	} else if (strncmp(devname, "eth", 3) && strncmp(devname, "usb", 3)) {
@@ -936,6 +996,8 @@ static int dump_regs(struct ethtool_drvinfo *info, struct ethtool_regs *regs)
 		return natsemi_dump_regs(info, regs);
 	if (!strncmp("e100", info->driver, ETHTOOL_BUSINFO_LEN))
 		return e100_dump_regs(info, regs);
+	if (!strncmp("amd8111e", info->driver, ETHTOOL_BUSINFO_LEN))
+		return amd8111e_dump_regs(info, regs);
 
 
 	fprintf(stdout, "Offset\tValue\n");
@@ -961,10 +1023,12 @@ static int dump_eeprom(struct ethtool_drvinfo *info, struct ethtool_eeprom *ee)
 		return tg3_dump_eeprom(info, ee);
 	}
 
-	fprintf(stdout, "Offset\tValue\n");
-	fprintf(stdout, "--------\t-----\n");
-	for (i = 0; i < ee->len; i++)
-		fprintf(stdout, "%02d\t0x%02x\n", i + ee->offset, ee->data[i]);
+	fprintf(stdout, "Offset\t\tValues\n");
+	fprintf(stdout, "------\t\t------");
+	for (i = 0; i < ee->len; i++) {
+		if(!(i%16)) fprintf(stdout, "\n0x%04x\t\t", i + ee->offset);
+		fprintf(stdout, "%02x ", ee->data[i]);
+	}
 	fprintf(stdout, "\n");
 	return 0;
 }
@@ -1091,15 +1155,17 @@ static int dump_coalesce(void)
 	return 0;
 }
 
-static int dump_checksum (int rx, int tx, int sg)
+static int dump_offload (int rx, int tx, int sg, int tso)
 {
 	fprintf(stdout,
 		"rx-checksumming: %s\n"
 		"tx-checksumming: %s\n"
-		"scatter-gather: %s\n",
+		"scatter-gather: %s\n"
+		"tcp segmentation offload: %s\n",
 		rx ? "on" : "off",
 		tx ? "on" : "off",
-		sg ? "on" : "off");
+		sg ? "on" : "off",
+		tso ? "on" : "off");
 
 	return 0;
 }
@@ -1133,6 +1199,8 @@ static int doit(void)
 		return do_nway_rst(fd, &ifr);
 	} else if (mode == MODE_GEEPROM) {
 		return do_geeprom(fd, &ifr);
+	} else if (mode == MODE_SEEPROM) {
+		return do_seeprom(fd, &ifr);
 	} else if (mode == MODE_TEST) {
 		return do_test(fd, &ifr);
 	} else if (mode == MODE_PHYS_ID) {
@@ -1149,10 +1217,10 @@ static int doit(void)
 		return do_gring(fd, &ifr);
 	} else if (mode == MODE_SRING) {
 		return do_sring(fd, &ifr);
-	} else if (mode == MODE_GSUM) {
-		return do_gsum(fd, &ifr);
-	} else if (mode == MODE_SSUM) {
-		return do_ssum(fd, &ifr);
+	} else if (mode == MODE_GOFFLOAD) {
+		return do_goffload(fd, &ifr);
+	} else if (mode == MODE_SOFFLOAD) {
+		return do_soffload(fd, &ifr);
 	} else if (mode == MODE_GSTATS) {
 		return do_gstats(fd, &ifr);
 	}
@@ -1358,12 +1426,12 @@ static int do_scoalesce(int fd, struct ifreq *ifr)
 	return 0;
 }
 
-static int do_gsum(int fd, struct ifreq *ifr)
+static int do_goffload(int fd, struct ifreq *ifr)
 {
 	struct ethtool_value eval;
-	int err, allfail = 1, rx = 0, tx = 0, sg = 0;
+	int err, allfail = 1, rx = 0, tx = 0, sg = 0, tso = 0;
 
-	fprintf(stdout, "RX/TX checksumming parameters for %s:\n", devname);
+	fprintf(stdout, "Offload parameters for %s:\n", devname);
 
 	eval.cmd = ETHTOOL_GRXCSUM;
 	ifr->ifr_data = (caddr_t)&eval;
@@ -1395,23 +1463,33 @@ static int do_gsum(int fd, struct ifreq *ifr)
 		allfail = 0;
 	}
 
+	eval.cmd = ETHTOOL_GTSO;
+	ifr->ifr_data = (caddr_t)&eval;
+	err = ioctl(fd, SIOCETHTOOL, ifr);
+	if (err)
+		perror("Cannot get device tcp segmentation offload settings");
+	else {
+		tso = eval.data;
+		allfail = 0;
+	}
+
 	if (allfail) {
-		fprintf(stdout, "no checksumming/SG info available\n");
+		fprintf(stdout, "no offload info available\n");
 		return 83;
 	}
 
-	return dump_checksum(rx, tx, sg);
+	return dump_offload(rx, tx, sg, tso);
 }
 
-static int do_ssum(int fd, struct ifreq *ifr)
+static int do_soffload(int fd, struct ifreq *ifr)
 {
 	struct ethtool_value eval;
 	int err, changed = 0;
 
-	if (csum_rx_wanted >= 0) {
+	if (off_csum_rx_wanted >= 0) {
 		changed = 1;
 		eval.cmd = ETHTOOL_SRXCSUM;
-		eval.data = (csum_rx_wanted == 1);
+		eval.data = (off_csum_rx_wanted == 1);
 		ifr->ifr_data = (caddr_t)&eval;
 		err = ioctl(fd, SIOCETHTOOL, ifr);
 		if (err) {
@@ -1420,10 +1498,10 @@ static int do_ssum(int fd, struct ifreq *ifr)
 		}
 	}
 
-	if (csum_tx_wanted >= 0) {
+	if (off_csum_tx_wanted >= 0) {
 		changed = 1;
 		eval.cmd = ETHTOOL_STXCSUM;
-		eval.data = (csum_tx_wanted == 1);
+		eval.data = (off_csum_tx_wanted == 1);
 		ifr->ifr_data = (caddr_t)&eval;
 		err = ioctl(fd, SIOCETHTOOL, ifr);
 		if (err) {
@@ -1432,10 +1510,10 @@ static int do_ssum(int fd, struct ifreq *ifr)
 		}
 	}
 
-	if (csum_sg_wanted >= 0) {
+	if (off_sg_wanted >= 0) {
 		changed = 1;
 		eval.cmd = ETHTOOL_SSG;
-		eval.data = (csum_sg_wanted == 1);
+		eval.data = (off_sg_wanted == 1);
 		ifr->ifr_data = (caddr_t)&eval;
 		err = ioctl(fd, SIOCETHTOOL, ifr);
 		if (err) {
@@ -1443,8 +1521,20 @@ static int do_ssum(int fd, struct ifreq *ifr)
 			return 86;
 		}
 	}
+	
+	if (off_tso_wanted >= 0) {
+		changed = 1;
+		eval.cmd = ETHTOOL_STSO;
+		eval.data = (off_tso_wanted == 1);
+		ifr->ifr_data = (caddr_t)&eval;
+		err = ioctl(fd, SIOCETHTOOL, ifr);
+		if (err) {
+			perror("Cannot set device tcp segmentation offload settings");
+			return 88;
+		}
+	}
 	if (!changed) {
-		fprintf(stdout, "no checksumming/SG settings changed\n");
+		fprintf(stdout, "no offload settings changed\n");
 	}
 
 	return 0;
@@ -1539,6 +1629,8 @@ static int do_sset(int fd, struct ifreq *ifr)
 				ecmd.phy_address = phyad_wanted;
 			if (xcvr_wanted != -1)
 				ecmd.transceiver = xcvr_wanted;
+			if (advertising_wanted != -1)
+				ecmd.advertising = advertising_wanted;
 
 			/* Try to perform the update. */
 			ecmd.cmd = ETHTOOL_SSET;
@@ -1637,9 +1729,14 @@ static int do_gregs(int fd, struct ifreq *ifr)
 	err = ioctl(fd, SIOCETHTOOL, ifr);
 	if (err < 0) {
 		perror("Cannot get register dump");
+		free(regs);
 		return 74;
 	}
-	dump_regs(&drvinfo, regs);
+	if(dump_regs(&drvinfo, regs) < 0) {
+		perror("Cannot dump registers");
+		free(regs);
+		return 75;
+	}
 	free(regs);
 
 	return 0;
@@ -1691,6 +1788,7 @@ static int do_geeprom(int fd, struct ifreq *ifr)
 	err = ioctl(fd, SIOCETHTOOL, ifr);
 	if (err < 0) {
 		perror("Cannot get EEPROM data");
+		free(eeprom);
 		return 74;
 	}
 	err = dump_eeprom(&drvinfo, eeprom);
@@ -1699,6 +1797,29 @@ static int do_geeprom(int fd, struct ifreq *ifr)
 	return err;
 }
 
+static int do_seeprom(int fd, struct ifreq *ifr)
+{
+	int err;
+	struct {
+		struct ethtool_eeprom eeprom;
+		u8 data;
+	} edata;
+
+	edata.eeprom.cmd = ETHTOOL_SEEPROM;
+	edata.eeprom.len = 1;
+	edata.eeprom.offset = seeprom_offset;
+	edata.eeprom.magic = seeprom_magic;
+	edata.data = seeprom_value;
+	ifr->ifr_data = (caddr_t)&edata.eeprom;
+	err = ioctl(fd, SIOCETHTOOL, ifr);
+	if (err < 0) {
+		perror("Cannot set EEPROM data");
+		return 87;
+	}
+
+	return err;
+}
+
 static int do_test(int fd, struct ifreq *ifr)
 {
 	int err;
@@ -1736,6 +1857,11 @@ static int do_test(int fd, struct ifreq *ifr)
 
 	strings = calloc(1, sizeof(*strings) +  
 			    drvinfo.testinfo_len * ETH_GSTRING_LEN);
+	if (!strings) {
+		perror("Cannot allocate memory for strings");
+		free(test);
+		return 73;
+	}
 	memset (strings->data, 0, drvinfo.testinfo_len * ETH_GSTRING_LEN);
 	strings->cmd = ETHTOOL_GSTRINGS;
 	strings->string_set = ETH_SS_TEST;
@@ -1809,6 +1935,8 @@ static int do_gstats(int fd, struct ifreq *ifr)
 	err = ioctl(fd, SIOCETHTOOL, ifr);
 	if (err < 0) {
 		perror("Cannot get stats strings information");
+		free(strings);
+		free(stats);
 		return 96;
 	}
 
@@ -1818,6 +1946,8 @@ static int do_gstats(int fd, struct ifreq *ifr)
 	err = ioctl(fd, SIOCETHTOOL, ifr);
 	if (err < 0) {
 		perror("Cannot get stats information");
+		free(strings);
+		free(stats);
 		return 97;
 	}
 
@@ -1831,6 +1961,8 @@ static int do_gstats(int fd, struct ifreq *ifr)
 		fprintf(stdout, "     %s: %llu\n",
 			s, stats->data[i]);
 	}
+	free(strings);
+	free(stats);
 
 	return 0;
 }
