@@ -281,14 +281,21 @@
 #define REG_SOPASS1			74
 #define REG_SOPASS2			75
 
+static void __print_intr(int d, int intr, const char *name,
+			 const char *s1, const char *s2)
+{
+	if ((d) & intr)
+		fprintf(stdout, "      %s Interrupt: %s\n", name, s1);
+	else if (s2)
+		fprintf(stdout, "      %s Interrupt: %s\n", name, s2);
+}
+
 #define PRINT_INTR(d, i, s1, s2) do { \
 	int intr = BIT_INTR_ ## i; \
-	char *name = NAME_INTR_ ## i; \
-	if ((d) & intr) \
-		fprintf(stdout, "      %s Interrupt: %s\n", name, s1); \
-	else if (s2) \
-		fprintf(stdout, "      %s Interrupt: %s\n", name, s2); \
+	const char *name = NAME_INTR_ ## i; \
+	__print_intr(d, intr, name, s1, s2); \
 } while (0)
+
 #define PRINT_INTRS(d, s1, s2) do { \
 	PRINT_INTR((d), RXOK, s1, s2); \
 	PRINT_INTR((d), RXDESC, s1, s2); \
